import abc
import os
import configparser

from hotsos.core.log import log


class ConfigBase(abc.ABC):
    """ Base class for config implementations. """
    def __init__(self, path):
        self.path = path

    @classmethod
    def squash_int_range(cls, ilist):
        """Takes a list of integers and squashes consecutive values into a
        string range.
        """
        irange = []
        rstart = None
        rprev = None

        sorted(ilist)
        for i, value in enumerate(ilist):
            if rstart is None:
                if i == (len(ilist) - 1):
                    irange.append(str(value))
                    break

                rstart = value

            if rprev is not None:
                if rprev != (value - 1):
                    if rstart == rprev:
                        irange.append(str(rstart))
                    else:
                        irange.append(f"{rstart}-{rprev}")
                        if i == (len(ilist) - 1):
                            irange.append(str(value))

                    rstart = value
                elif i == (len(ilist) - 1):
                    irange.append(f"{rstart}-{value}".format(rstart, value))
                    break

            rprev = value

        return ','.join(irange)

    @classmethod
    def expand_value_ranges(cls, ranges):
        """
        Takes a string containing ranges of values such as 1-3 and 4,5,6,7 and
        expands them into a single list.
        """
        if not ranges:
            return ranges

        expanded = []
        ranges = ranges.split(',')
        for subrange in ranges:
            # expand ranges
            subrange = subrange.partition('-')
            if subrange[1] == '-':
                expanded += range(int(subrange[0]), int(subrange[2]) + 1)
            else:
                for val in subrange[0].split():
                    expanded.append(int(val))

        return sorted(expanded)

    @property
    def exists(self):
        if os.path.exists(self.path):
            return True

        return False

    @abc.abstractmethod
    def get(self, key, section=None, expand_to_list=False):
        """ Get a config value. """


class IniConfigBase(ConfigBase):
    """ INI-style config file implementation. """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.config = None
        self._load()

    def __bool__(self):
        return self.config is not None

    @staticmethod
    def bool_str(val):
        if val.lower() == "true":
            return True
        if val.lower() == "false":
            return False

        return val

    @classmethod
    def post_processing(cls, v, expand_to_list=False):
        # Sanitize the string and perform
        # boolean type conversion if needed
        if isinstance(v, str):
            v = v.strip()
            for char in ["'", '"']:
                v = v.strip(char)
            v = cls.bool_str(v)

        if expand_to_list:
            return cls.expand_value_ranges(v)

        return v

    @property
    def all_sections(self):
        return self.config.sections() + ["DEFAULT"]

    @property
    def all_keys(self):
        return [x for option in self.config.items()
                for x in list(option[1].keys())]

    def get(self, key, section=None, expand_to_list=False):
        """ If section is None, then search all sections and return first
        match. """
        log.debug("ini read call -> %s:%s", section, key)

        if not self.config:
            return None

        # 1: Section is specified.
        if section is not None:
            # Look for section name, case-insensitive
            for sect in self.all_sections:
                if section.upper() == sect.upper():
                    return self.post_processing(
                        self.config.get(sect, key, fallback=None),
                        expand_to_list
                    )
            return None

        # 2: No section specified.
        # The default section is not included in the section
        # set so append it
        for sec in self.all_sections:
            v = self.config.get(sec, key, fallback=None)
            log.debug("ini read value -> %s:%s = %s", section, key, v)
            if v:
                return self.post_processing(v, expand_to_list)

        return None

    def _load(self):
        if not self.exists:
            log.debug("config file %s does not exist", self.path)
            self.config = None
            return

        self.config = configparser.ConfigParser(strict=False)
        try:
            parsed = self.config.read(self.path)
        except configparser.ParsingError as exc:
            log.warning("failed to read config file '%s': %s", self.path, exc)
            self.config = None
        else:
            if self.path not in parsed:
                log.warning("cannot parse config file '%s'", self.path)
                self.config = None


class GenericIniConfig(IniConfigBase):
    """ Generic INI file config implementation. """



from hotsos.core.plugins.openstack.common import (
    OpenstackBase,
    OpenStackChecks,
)
from hotsos.core.plugins.openstack.sunbeam import SunbeamInfo
from hotsos.core.plugintools import (
    summary_entry,
    get_min_available_entry_index,
)


class SunbeamStatus(OpenstackBase, OpenStackChecks):
    """ Get information from Sunbeam to display. """
    summary_part_index = 15

    @staticmethod
    @summary_entry('sunbeam', get_min_available_entry_index() + 10)
    def summary_sunbeam():
        sunbeam = SunbeamInfo()
        if sunbeam.pods:
            return {'pods': sunbeam.pods,
                    'statefulsets': sunbeam.statefulsets}

        return None



from .common import (
    ApacheInfo,
    OpenstackBase,
    OpenStackChecks,
    OpenstackEventHandlerBase,
    OpenstackEventCallbackBase,
)
from .search_exprs import (
    HTTPStatusExprs,
    HTTPRequestExprs,
)
from .openstack import OpenstackConfig

__all__ = [
    ApacheInfo.__name__,
    OpenstackBase.__name__,
    OpenStackChecks.__name__,
    OpenstackConfig.__name__,
    OpenstackEventHandlerBase.__name__,
    OpenstackEventCallbackBase.__name__,
    HTTPStatusExprs.__name__,
    HTTPRequestExprs.__name__,
    ]



# Must include all summary plugin handlers that need to be autoregister
from . import (  # noqa: F401
    juju,
    kernel,
    kubernetes,
    landscape,
    lxd,
    maas,
    microcloud,
    mysql,
    openstack,
    openvswitch,
    pacemaker,
    rabbitmq,
    sosreport,
    storage,
    system,
    vault
)



import os
import re

from hotsos.core.config import HotSOSConfig
from hotsos.core import host_helpers


class KernelConfig(host_helpers.ConfigBase):
    """ Kernel configuration. """
    def __init__(self, *args, **kwargs):
        path = os.path.join(HotSOSConfig.data_root, "proc/cmdline")
        super().__init__(path=path, *args, **kwargs)
        self._cfg = {}
        self._load()

    def get(self, key, section=None, expand_to_list=False):
        value = self._cfg.get(key)
        if expand_to_list and value is not None:
            return self.expand_value_ranges(value)

        return value

    def _load(self):
        """
        Currently only supports extracting isolcpus

        TODO: make more general
        """
        if not self.exists:
            return

        with open(self.path, encoding='utf-8') as fd:
            for line in fd:
                expr = r'.*\s+isolcpus=([0-9\-,]+)\s*.*'
                ret = re.compile(expr).match(line)
                if ret:
                    self._cfg["isolcpus"] = ret[1]
                    break


class SystemdConfig(host_helpers.IniConfigBase):
    """Systemd configuration."""
    def __init__(self, *args, **kwargs):
        path = os.path.join(HotSOSConfig.data_root, "etc/systemd/system.conf")
        super().__init__(path=path, *args, **kwargs)

    @property
    def cpuaffinity_enabled(self):
        if self.get('CPUAffinity'):
            return True

        return False



import os
import re
from functools import cached_property

from hotsos.core.config import HotSOSConfig
from hotsos.core.host_helpers import CLIHelperFile
from hotsos.core.log import log
from hotsos.core.ycheck.engine.properties.common import (
    ImportHelper,
    YPropertyOverrideBase,
)


class YPropertyInputBase(ImportHelper):
    """ Base class for input property implementations. """
    @property
    def options(self):
        defaults = {'disable-all-logs': False,
                    'args': [],
                    'kwargs': {},
                    'args-callback': None}

        if isinstance(self.content, dict):  # pylint: disable=E1101
            _options = self.content.get('options', defaults)  # noqa, pylint: disable=E1101
            defaults.update(_options)

        return defaults

    @property
    def command(self):
        if not isinstance(self.content, dict):  # pylint: disable=E1101
            return None

        return self.content.get('command')  # pylint: disable=E1101

    def expand_paths(self, paths):
        """
        Expand paths to absolute using prefix that exists.

        @param paths: list of paths we want to use as input. Note that paths
                      must be relative to data_root.
        """
        _paths = []
        for path in paths:
            # First check if a PathFinder is used
            if re.match(r'^[\w.]+:', path):
                plugin, _, subpath = path.partition(':')
                importpath = f'hotsos.core.plugins.{plugin}.PathFinder'
                # NOTE: we don't bother to supply context/cache. The pathfinder
                # itself should support affinity for paths that match.
                path = self.get_cls(None, importpath)().resolve(subpath)
                if not path:
                    log.debug("PathFinder did not identify a path for %s",
                              subpath)
                    continue

            path = os.path.join(HotSOSConfig.data_root, path)
            if (HotSOSConfig.use_all_logs and not
                    self.options['disable-all-logs']):
                path = f"{path}*"

            _paths.append(path)

        return _paths

    @property
    def path(self):
        raise AttributeError("do not call this directly")

    @cached_property
    def paths(self):
        fs_path = None
        if isinstance(self.content, (str, list)):  # pylint: disable=E1101
            fs_path = self.content  # pylint: disable=E1101
        else:
            fs_path = self.content.get('path')  # pylint: disable=E1101

        if fs_path:
            _paths = []
            if isinstance(fs_path, list):
                for path in fs_path:
                    _paths.append(path)
            else:
                _paths.append(fs_path)

            return self.expand_paths(_paths)

        if self.command:
            cmd_tmp_path = self.cache.cmd_tmp_path  # pylint: disable=E1101
            if cmd_tmp_path:
                return [cmd_tmp_path]

            args_callback = self.options['args-callback']
            if args_callback:
                args, kwargs = self.get_method(args_callback)
            else:
                args = self.options['args']
                kwargs = self.options['kwargs']

            with CLIHelperFile(delete_temp=False) as cli:
                outfile = getattr(cli, self.command)(*args, **kwargs)
                self.cache.set('cmd_tmp_path', outfile)  # noqa, pylint: disable=E1101
                return [outfile]

        log.debug("no input provided")
        return None


class YPropertyInput(YPropertyOverrideBase, YPropertyInputBase):
    """
    Input property. Provides support for defining input such a file or command
    output.

    This is typically used in conjunction with the search property.
    """
    override_keys = ['input']
    # We want to be able to use this property both on its own and as a member
    # of other mapping properties e.g. Checks. The following setting enables
    # this.
    override_auto_implicit_member = False



import os
import tempfile
from collections import UserList

from hotsos.core.host_helpers.cli.common import BinCmd, FileCmd
from hotsos.core.host_helpers.exceptions import SourceNotFound


CephAliases = ['microceph.']


class CephJSONFileCmd(FileCmd):
    """
    Some ceph commands that use --format json have some extra text added to the
    end of the file (typically from stderr) which causes it to be invalid json
    so we have to strip that final line before decoding the contents.
    """
    def __init__(self, *args, first_line_filter=None, last_line_filter=None,
                 **kwargs):
        super().__init__(*args, **kwargs)
        if first_line_filter or last_line_filter:
            self.register_hook('pre-exec', self.format_json_contents)
            self.register_hook('post-exec', self.cleanup)
            self.orig_path = None
            self.first_line_filter = first_line_filter
            self.last_line_filter = last_line_filter

    def format_json_contents(self, *_args, **_kwargs):
        if not os.path.exists(self.path):
            raise SourceNotFound(self.path)

        with open(self.path, encoding='utf-8') as f:
            lines = f.readlines()

        if self.first_line_filter:
            line_filter = self.first_line_filter
        else:
            line_filter = self.last_line_filter

        if lines and lines[-1].startswith(line_filter):
            lines = lines[:-1]
            with tempfile.NamedTemporaryFile(mode='w+t', delete=False) as tmp:
                tmp.write(''.join(lines))
                tmp.close()
                self.orig_path = self.path
                self.path = tmp.name

    def cleanup(self, output, **_kwargs):
        """
        @param output: CmdOutput object
        """
        if self.orig_path:
            os.unlink(self.path)
            self.path = self.orig_path
            self.orig_path = None

        return output


class CephHealthDetailCommands(UserList):
    """ Generate ceph heath detail command variants. """

    def __init__(self):
        prefixes = [""] + CephAliases
        # binary
        cmds = [BinCmd(f'{prefix}ceph health detail --format json-pretty',
                       json_decode=True) for prefix in prefixes]
        # file-based
        # sosreport < 4.2
        cmds.append(FileCmd('sos_commands/ceph/json_output/'
                            'ceph_health_detail_--format_json-pretty',
                            json_decode=True))
        # sosreport >= 4.2
        cmds.extend([FileCmd('sos_commands/ceph_mon/json_output/'
                             f'{prefix}ceph_health_detail_'
                             '--format_json-pretty', json_decode=True)
                     for prefix in prefixes])
        super().__init__(cmds)


class CephMonDumpCommands(UserList):
    """ Generate ceph mon dump command variants. """

    def __init__(self):
        prefixes = [""] + CephAliases
        # binary
        cmds = [BinCmd(f'{prefix}ceph mon dump --format json-pretty',
                       json_decode=True) for prefix in prefixes]
        # file-based
        # sosreport < 4.2
        cmds.append(CephJSONFileCmd('sos_commands/ceph/json_output/'
                                    'ceph_mon_dump_--format_json-pretty',
                                    json_decode=True,
                                    first_line_filter='dumped monmap epoch',
                                    last_line_filter='dumped monmap epoch'))
        # sosreport >= 4.2
        cmds.extend([CephJSONFileCmd('sos_commands/ceph_mon/json_output/'
                                     f'{prefix}ceph_mon_dump_'
                                     '--format_json-pretty',
                                     json_decode=True,
                                     first_line_filter='dumped monmap epoch',
                                     last_line_filter='dumped monmap epoch')
                     for prefix in prefixes])
        super().__init__(cmds)


class CephOSDDumpCommands(UserList):
    """ Generate ceph osd dump command variants. """

    def __init__(self):
        prefixes = [""] + CephAliases
        # binary
        cmds = [BinCmd(f'{prefix}ceph osd dump --format json-pretty',
                       json_decode=True)
                for prefix in prefixes]
        # file-based
        # sosreport < 4.2
        cmds.append(FileCmd('sos_commands/ceph/json_output/'
                            'ceph_osd_dump_--format_json-pretty',
                            json_decode=True))
        # sosreport >= 4.2
        cmds.extend([FileCmd('sos_commands/ceph_mon/json_output/'
                             f'{prefix}ceph_osd_dump_--format_json-pretty',
                             json_decode=True)
                     for prefix in prefixes])
        super().__init__(cmds)


class CephDFCommands(UserList):
    """ Generate ceph df command variants. """

    def __init__(self):
        prefixes = [""] + CephAliases
        # binary
        cmds = [BinCmd(f'{prefix}ceph df --format json-pretty',
                       json_decode=True) for prefix in prefixes]
        # file-based
        # sosreport < 4.2
        cmds.append(FileCmd('sos_commands/ceph/json_output/'
                            'ceph_df_--format_json-pretty',
                            json_decode=True))
        # sosreport >= 4.2
        cmds.extend([FileCmd('sos_commands/ceph_mon/json_output/'
                             f'{prefix}ceph_df_--format_json-pretty',
                             json_decode=True)
                     for prefix in prefixes])
        super().__init__(cmds)


class CepOSDDFTreeCommands(UserList):
    """ Generate ceph df tree command variants. """

    def __init__(self):
        prefixes = [""] + CephAliases
        # binary
        cmds = [BinCmd(f'{prefix}ceph osd df tree --format json-pretty',
                       json_decode=True) for prefix in prefixes]
        # file-based
        # sosreport < 4.2
        cmds.append(FileCmd('sos_commands/ceph/json_output/'
                            'ceph_osd_df_tree_--format_json-pretty',
                            json_decode=True))
        # sosreport >= 4.2
        cmds.extend([FileCmd('sos_commands/ceph_mon/json_output/'
                             f'{prefix}ceph_osd_df_tree_--format_json-pretty',
                             json_decode=True)
                     for prefix in prefixes])
        super().__init__(cmds)


class CephOSDCrushDumpCommands(UserList):
    """ Generate ceph osd crush dump command variants. """

    def __init__(self):
        prefixes = [""] + CephAliases
        # binary
        cmds = [BinCmd(f'{prefix}ceph osd crush dump',
                       json_decode=True) for prefix in prefixes]
        # file-based
        # sosreport < 4.2
        cmds.append(FileCmd('sos_commands/ceph/ceph_osd_crush_dump',
                            json_decode=True))
        # sosreport >= 4.2
        cmds.extend([FileCmd('sos_commands/ceph_mon/'
                             f'{prefix}ceph_osd_crush_dump',
                             json_decode=True)
                     for prefix in prefixes])
        super().__init__(cmds)


class CephPGDumpCommands(UserList):
    """ Generate ceph pg dump command variants. """

    def __init__(self):
        prefixes = [""] + CephAliases
        # binary
        cmds = [BinCmd(f'{prefix}ceph pg dump --format json-pretty',
                       json_decode=True)
                for prefix in prefixes]
        # file-based
        # sosreport < 4.2
        cmds.append(CephJSONFileCmd('sos_commands/ceph/json_output/'
                                    'ceph_pg_dump_--format_json-pretty',
                                    json_decode=True,
                                    first_line_filter='dumped all',
                                    last_line_filter='dumped all'))
        # sosreport >= 4.2
        cmds.extend([CephJSONFileCmd('sos_commands/ceph_mon/json_output/'
                                     f'{prefix}ceph_pg_dump_'
                                     '--format_json-pretty',
                                     json_decode=True,
                                     first_line_filter='dumped all',
                                     last_line_filter='dumped all')
                     for prefix in prefixes])
        super().__init__(cmds)


class CephStatusCommands(UserList):
    """ Generate ceph status command variants. """

    def __init__(self):
        prefixes = [""] + CephAliases
        # binary
        cmds = [BinCmd(f'{prefix}ceph status --format json-pretty',
                       json_decode=True) for prefix in prefixes]
        # file-based
        # sosreport < 4.2
        cmds.append(FileCmd('sos_commands/ceph/json_output/'
                            'ceph_status_--format_json-pretty',
                            json_decode=True))
        # sosreport >= 4.2
        cmds.extend([FileCmd('sos_commands/ceph_mon/json_output/'
                             f'{prefix}ceph_status_--format_json-pretty',
                             json_decode=True)
                     for prefix in prefixes])
        super().__init__(cmds)


class CephVersionsCommands(UserList):
    """ Generate ceph versions command variants. """

    def __init__(self):
        prefixes = [""] + CephAliases
        # binary
        cmds = [BinCmd(f'{prefix}ceph versions') for prefix in prefixes]
        # file-based
        # sosreport < 4.2
        cmds.append(FileCmd('sos_commands/ceph/ceph_versions'))
        # sosreport >= 4.2
        cmds.extend([FileCmd('sos_commands/ceph_mon/'
                             f'{prefix}ceph_versions')
                     for prefix in prefixes])
        super().__init__(cmds)


class CephVolumeLVMListCommands(UserList):
    """ Generate ceph-volume lvm list command variants. """

    def __init__(self):
        prefixes = [""] + CephAliases
        # binary
        cmds = [BinCmd(f'{prefix}ceph-volume lvm list') for prefix in prefixes]
        # file-based
        # sosreport < 4.2
        cmds.append(FileCmd('sos_commands/ceph/ceph-volume_lvm_list'))
        # sosreport >= 4.2
        cmds.extend([FileCmd('sos_commands/ceph_osd/'
                             f'{prefix}ceph-volume_lvm_list')
                     for prefix in prefixes])
        super().__init__(cmds)


class CephReportCommands(UserList):
    """ Generate ceph XXX command variants. """

    def __init__(self):
        prefixes = [""] + CephAliases
        # binary
        cmds = [BinCmd(f'{prefix}ceph report', json_decode=True)
                for prefix in prefixes]
        # file-based
        # sosreport < 4.2
        cmds.append(CephJSONFileCmd('sos_commands/ceph/ceph_report',
                                    json_decode=True,
                                    first_line_filter='report',
                                    last_line_filter='report'))
        # sosreport >= 4.2
        cmds.extend([CephJSONFileCmd('sos_commands/ceph_mon/'
                                     f'{prefix}ceph_report',
                                     json_decode=True,
                                     first_line_filter='report',
                                     last_line_filter='report')
                     for prefix in prefixes])
        super().__init__(cmds)


class CephMgrModuleLsCommands(UserList):
    """ Generate ceph XXX command variants. """

    def __init__(self):
        prefixes = [""] + CephAliases
        # binary
        cmds = [BinCmd(f'{prefix}ceph mgr module ls',
                       json_decode=True) for prefix in prefixes]
        # file-based
        # sosreport < 4.2
        cmds.append(CephJSONFileCmd('sos_commands/ceph/ceph_mgr_module_ls',
                                    json_decode=True))
        # sosreport >= 4.2
        cmds.extend([CephJSONFileCmd('sos_commands/ceph_mon/'
                                     f'{prefix}ceph_mgr_module_ls',
                                     json_decode=True)
                     for prefix in prefixes])
        super().__init__(cmds)



from hotsos.core.plugins.vault import VaultChecks


class VaultSummary(VaultChecks):
    """ Implementation of Vault summary. """
    summary_part_index = 0

    # REMINDER: Common entries are implemented in
    #           plugintools.ApplicationSummaryBase. Only customisations are
    #           implemented here. See
    #           plugintools.get_min_available_entry_index() for an explanation
    #           on how entry indices are managed.

    # NOTE: no custom entries defined here yet so plugin will only use
    #       defaults.
